/********************************************
fin_exec.c

libmawk changes (C) 2009-2013, Tibor 'Igor2' Palinkas;
based on mawk code coming with the below copyright:

copyright 1991, 1992.  Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/

#ifndef _POSIX_SOURCE
/* needed on some systems to get strdup() */
#define _POSIX_SOURCE
#endif
#include <stdlib.h>
#include "mawk.h"
#include "fin.h"
#include "memory.h"
#include "bi_vars.h"
#include "field.h"
#include "symtype.h"
#include "scan.h"
#include "vio.h"
#include "init.h"
#include "vars.h"
#include <string.h>

#ifndef	  NO_FCNTL_H
#include <fcntl.h>
#endif

/* main_fin is a handle to the main input stream
   == 0	 never been opened   */

static void set_main_to_stdin(mawk_state_t * MAWK)
{
	cell_destroy(FILENAME);
	FILENAME->type = C_STRING;
	FILENAME->ptr = (PTR) mawk_new_STRING(MAWK, "-");
	cell_destroy(FNR);
	FNR->type = C_NUM;
	FNR->dval = NUM_ZERO;
	MAWK->rt_fnr = 0;
	MAWK->main_fin = mawk_fin_fd0_open(MAWK, 1);
}

/* get the next command line file open */
static FIN *next_main(mawk_state_t *MAWK, int open_flag)
{
	/* open_flag: called by mawk_FINopen_main() if on */
	register CELL *cp;
	CELL argc;										/* copy of ARGC */
	CELL c_argi;									/* cell copy of argi */
	CELL argval;									/* copy of ARGV[c_argi] */


	argval.type = C_NOINIT;
	c_argi.type = C_NUM;

	if (MAWK->main_fin)
		mawk_FINclose(MAWK, MAWK->main_fin);
		MAWK->main_fin = NULL;
	/* FILENAME and FNR don't change unless we really open
	   a new file */

	/* make a copy of ARGC to avoid side effect */
	if (mawk_cellcpy(MAWK, &argc, ARGC)->type != C_NUM)
		mawk_cast1_to_d(MAWK, &argc);

	while (MAWK->argi < argc.dval) {
		c_argi.dval = MAWK->argi;
		MAWK->argi += NUM_ONE;

		if (!(cp = mawk_array_find(MAWK, MAWK->Argv, &c_argi, NO_CREATE)))
			continue;									/* its deleted */

		/* make a copy so we can mawk_cast w/o side effect */
		cell_destroy(&argval);
		cp = mawk_cellcpy(MAWK, &argval, cp);
		if (cp->type < C_STRING)
			mawk_cast1_to_s(MAWK, cp);
		if (string(cp)->len == 0)
			continue;
		/* file argument is "" */

		/* it might be a command line assignment */
		if (mawk_is_cmdline_assign(MAWK, string(cp)->str))
			continue;

		/* try to open it -- we used to continue on failure,
		   but posix says we should quit */
		if (!(MAWK->main_fin = mawk_FINopen(MAWK, string(cp)->str, 1))) {
			mawk_errmsg(MAWK, errno, "cannot open %s", string(cp)->str);
			mawk_exitval(MAWK, 2, NULL);
		}

		/* success -- set FILENAME and FNR */
		cell_destroy(FILENAME);
		mawk_cellcpy(MAWK, FILENAME, cp);
		free_STRING(string(cp));
		cell_destroy(FNR);
		FNR->type = C_NUM;
		FNR->dval = NUM_ZERO;
		MAWK->rt_fnr = 0;

		return MAWK->main_fin;
	}
	/* failure */
	cell_destroy(&argval);

	if (open_flag) {
		/* all arguments were null or assignment */
		set_main_to_stdin(MAWK);
		return MAWK->main_fin;
	}

	/* real failure */
	{
		/* this is how we mark EOF on main_fin  */
		static const char dead_buff = 0;
		static const FIN dead_main = { NULL, (void *)&dead_buff, (void *)&dead_buff,
			1, EOF_FLAG, NULL, 0
		};


		return MAWK->main_fin = (void *)&dead_main;
		/* since MAIN_FLAG is not set, mawk_FINgets won't call next_main() */
	}
}




/* this gets called once to get the input stream going.
   It is called after the execution of the BEGIN block
   unless there is a getline inside BEGIN {}
*/
void mawk_FINopen_main(mawk_state_t * MAWK)
{
	CELL argc;

	mawk_cellcpy(MAWK, &argc, ARGC);
	if (argc.type != C_NUM)
		mawk_cast1_to_d(MAWK, &argc);

	if (argc.dval == NUM_ONE)
		set_main_to_stdin(MAWK);
	else
		next_main(MAWK, 1);
}

int mawk_is_cmdline_assign(mawk_state_t *MAWK, char *s)
{
	register char *p;
	int c;
	CELL *cp;
	unsigned len;
	CELL *fp = (CELL *) 0;				/* ditto */

	if (MAWK->scan_code[*(unsigned char *) s] != SC_IDCHAR)
		return 0;

	p = s + 1;
	while ((c = MAWK->scan_code[*(unsigned char *) p]) == SC_IDCHAR || c == SC_DIGIT)
		p++;

	if (*p != '=')
		return 0;

	*p = 0;

	cp = mawk_create_var(MAWK, s, &fp);
	if (cp == NULL) {
		mawk_rt_error(MAWK, "cannot command line assign to %s\n\ttype clash or keyword", s);
	}

	/* we need to keep ARGV[i] intact */
	*p++ = '=';
	len = strlen(p) + 1;
	/* posix says escape sequences are on from command line */
	p = mawk_rm_escape(MAWK, strcpy((char *) zmalloc(MAWK, len), p));
	cp->ptr = (PTR) mawk_new_STRING(MAWK, p);
	zfree(MAWK, p, len);
	mawk_check_strnum(MAWK, cp);	/* sets cp->type */
	if (fp) {											/* move it from cell to pfield[] */
		mawk_field_assign(MAWK, fp, cp);
		free_STRING(string(cp));
	}
	return 1;
}

static char *enlarge_fin_buffer(mawk_state_t *MAWK, FIN *fin)
{
	unsigned r;
	unsigned oldsize = fin->nbuffs * BUFFSZ + 1;

	fin->buffp = fin->buff = (char *) mawk_zrealloc(MAWK, fin->buff, oldsize, oldsize + BUFFSZ);
	fin->nbuffs++;

	r = mawk_fillbuff(MAWK, fin, fin->buff + (oldsize - 1), BUFFSZ);
	if (r < BUFFSZ)
		fin->flags |= EOF_FLAG;

	return fin->buff;
}

/* return one input record as determined by RS,
   from input file (FIN)  fin
*/

char *mawk_FINgets(mawk_state_t *MAWK, FIN *fin, unsigned *len_p)
{
	register char *p, *q;
	unsigned match_len;
	unsigned r;

restart:

	if (MAWK->do_exit)
		return NULL;

	if (!(p = fin->buffp)[0]) {		/* need a refill */
		if (fin->flags & EOF_FLAG) {
			if (fin->flags & MAIN_FLAG) {
				fin = next_main(MAWK, 0);
				goto restart;
			}
			else {
				*len_p = 0;
				return (char *) 0;
			}
		}
		
		if ((fin->artificial_size <= 0) && mawk_vio_fgetsable(MAWK, fin->vf)) {
			/* line buffering */
			if (!mawk_vio_fgets(MAWK, fin->vf, fin->buff, BUFFSZ + 1)) {
				fin->flags |= EOF_FLAG;
				fin->buff[0] = 0;
				fin->buffp = fin->buff;
				goto restart;						/* might be main_fin */
			}
			else {										/* return this line */

				/* mawk_find eol */
				p = fin->buff;
				while (*p != '\n' && *p != 0)
					p++;

				*p = 0;
				*len_p = p - fin->buff;
				fin->buffp = p;
				return fin->buff;
			}
		}
		else {
			/* block buffering */
			r = mawk_fillbuff(MAWK, fin, fin->buff, fin->nbuffs * BUFFSZ);
			if (r == 0) {
				fin->flags |= EOF_FLAG;
				fin->buffp = fin->buff;
				goto restart;						/* might be main */
			}
			else if (r == mawk_FIN_nomore)
				return (char *)mawk_FIN_nomore;

/* we will detect eof next mawk_FINgets anyway
			else if (r < fin->nbuffs * BUFFSZ) {
				fin->flags |= EOF_FLAG;
			}*/

			p = fin->buffp = fin->buff;

			if (fin->flags & START_FLAG) {
				fin->flags &= ~START_FLAG;
				if (MAWK->rs_shadow.type == SEP_MLR) {
					/* trim blank lines from front of file */
					while (*p == '\n')
						p++;
					fin->buffp = p;
					if (*p == 0)
						goto restart;
				}
			}
		}
	}

retry:

	if (MAWK->do_exit)
		return NULL;


	switch (MAWK->rs_shadow.type) {
	case SEP_CHAR:
		q = strchr(p, MAWK->rs_shadow.c);
		match_len = 1;
		break;

	case SEP_STR:
		q = mawk_str_str(p, ((STRING *) MAWK->rs_shadow.ptr)->str, match_len = ((STRING *) MAWK->rs_shadow.ptr)->len);
		break;

	case SEP_MLR:
	case SEP_RE:
		q = mawk_re_pos_match(MAWK, p, MAWK->rs_shadow.ptr, &match_len);
		/* if the match is at the end, there might still be
		   more to match in the file */
		if (q && q[match_len] == 0 && !(fin->flags & EOF_FLAG))
			q = (char *) 0;
		break;

	default:
		mawk_bozo(MAWK, "type of rs_shadow");
	}

	if (q) {
		/* the easy and normal case */
		*q = 0;
		*len_p = q - p;
		fin->buffp = q + match_len;
		return p;
	}

	if (fin->flags & EOF_FLAG) {
		/* last line without a record terminator */
		*len_p = r = strlen(p);
		fin->buffp = p + r;

		if (MAWK->rs_shadow.type == SEP_MLR && fin->buffp[-1] == '\n' && r != 0) {
			(*len_p)--;
			*--fin->buffp = 0;
		}
		return p;
	}

	if (p == fin->buff) {
		/* current record is too big for the input buffer, grow buffer */
		p = enlarge_fin_buffer(MAWK, fin);
	}
	else {
		/* move a partial line to front of buffer and try again */
		unsigned rr;

		p = (char *) memcpy(fin->buff, p, r = strlen(p));
		q = p + r;
		rr = fin->nbuffs * BUFFSZ - r;

		if ((r = mawk_fillbuff(MAWK, fin, q, rr)) < rr) {
			if (r == mawk_FIN_nomore)
				return (char *)mawk_FIN_nomore;
			fin->flags |= EOF_FLAG;
		}
	}
	if (MAWK->do_exit)
		return NULL;
	goto retry;
}



