#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "libmawk.h"
#include "debug.h"
#include "memory.h"
#include "sizes.h"
#include "fin.h"
#include "num.h"
#include "vio.h"
#include "vars.h"

mawk_state_t *libmawk_initialize_stage1(int argc, char *argv[])
{
	mawk_state_t *m;
	char *argv_dummy[] = {
		"(null)",
		"-f",
		"/dev/null",
		NULL
	};

	if (argv == NULL) {
		argc = 3;
		argv = argv_dummy;
	}

	/* load and parse */
	m = mawk_initialize(argc, argv);

	if (m == NULL)
		return NULL;

	m->dont_close_std = 1;
	m->separate_begin = 1;
	m->suppress_undefined_function_warning = 1;
	return m;
}

mawk_state_t *libmawk_initialize_stage2(mawk_state_t *m)
{
	if (*m->ps.buffp == '\0')
		strcpy((char *)m->ps.buffp, "BEGIN {}\n");
	mawk_parse(m) ;
	m->mawk_state = EXECUTION ;

	/* run BEGIN blocks */
	m->main_fin = mawk_fin_fd0_open(m, 1);

#warning TODO: mark fin as a vritual file
/*	m->main_fin->fp = NULL;
	m->main_fin->fd = -2;*/
	
	m->main_fin->artificial_buff = NULL;
	m->main_fin->artificial_size = 0;
	return m;
}

mawk_state_t *libmawk_initialize_stage3(mawk_state_t *m)
{
	if (m->debug_symbols)
		mawk_debug_callstack_push(m, &mawk_debug_begin);

	mawk_execute(m, m->execution_start, m->eval_stack - 1, 0) ;

	if (m->debug_symbols)
		mawk_debug_callstack_pop(m);

	return m;
}

mawk_state_t *libmawk_initialize(int argc, char *argv[])
{
	mawk_state_t *m;

	m = libmawk_initialize_stage1(argc, argv);

	if (m == NULL)
		return NULL;

	m = libmawk_initialize_stage2(m);
	m = libmawk_initialize_stage3(m);

	return m;
}


void libmawk_run_main(mawk_state_t *m)
{
	if (m->debug_symbols)
		mawk_debug_callstack_push(m, &mawk_debug_main);

	mawk_execute(m, m->main_start, m->eval_stack - 1, 0);

	if (m->debug_symbols)
		mawk_debug_callstack_pop(m);
}

void libmawk_uninitialize(mawk_state_t *m)
{
	INST exit0 = {_EXIT0};
	mawk_execute(m, &exit0, m->eval_stack - 1, 0);

	mawk_FINclose(m, m->main_fin);

	mawk_uninitialize(m);
}

/* callback from mawk */
void mawk_exit_(mawk_state_t *MAWK, int x)
{
	mawk_close_out_pipes(MAWK) ;		 /* no effect, if no out pipes */
	/* we do not really exit */
	MAWK->final_exit_code = MAWK->rt_exit_code != 0 ? MAWK->rt_exit_code : x;
}

void libmawk_append_ninput(mawk_state_t *m, const char *input, int len)
{
	int end;

	if (m->main_fin->artificial_buff == NULL) {
		m->main_fin->artificial_buff = malloc(len);
		memcpy(m->main_fin->artificial_buff, input, len);
		m->main_fin->artificial_size = len;
	}
	else {
		end = m->main_fin->artificial_size;
		m->main_fin->artificial_buff = realloc(m->main_fin->artificial_buff, end+len+1);
		memcpy(m->main_fin->artificial_buff+end, input, len);
		m->main_fin->artificial_size += len;
	}
#warning TODO: temp hack until real vio instead of artificial buff
	/* detach stdin */
	m->main_fin->vf = NULL;
}

void libmawk_append_input(mawk_state_t *m, const char *input_str)
{
	libmawk_append_ninput(m, input_str, strlen(input_str));
}


CELL *libmawk_set_cellv(mawk_state_t *MAWK, CELL *cell, const char argtype, va_list *ap)
{
	char *s;
	int i;
	mawk_num_t d;

	switch(argtype) {
		case 's':
			s = va_arg((*ap), char *);
			cell->type = C_STRING ;
			cell->ptr = mawk_new_STRING(MAWK, s);
			break;
		case 'd':
			i = va_arg((*ap), int);
			cell->type = C_NUM ;
			cell->dval = i;
			break;
#ifndef MAWK_NO_FLOAT
		case 'f':
			d = va_arg((*ap), double);
			cell->type = C_NUM;
			cell->dval = d;
			break;
#endif
		default:
			return NULL;
	}
	return cell;
}

CELL *libmawk_set_cellp(mawk_state_t *MAWK, CELL *cell, const char argtype, void *argp)
{
	char *s;
	int i;
	mawk_num_t d;

	switch(argtype) {
		case 's':
			s = (char *)argp;
			cell->type = C_STRING ;
			cell->ptr = mawk_new_STRING(MAWK, s);
			break;
		case 'd':
			i = *(int *)argp;
			cell->type = C_NUM ;
			cell->dval = i;
			break;
#ifndef MAWK_NO_FLOAT
		case 'f':
			d = *(double *)argp;
			cell->type = C_NUM ;
			cell->dval = d;
			break;
#endif
		default:
			return NULL;
	}
	return cell;
}

CELL *libmawk_set_cell(mawk_state_t *MAWK, CELL *cell, const char argtype, ...)
{
	va_list ap;
	CELL *ret;

	va_start(ap, argtype);
	ret = libmawk_set_cellv(MAWK, cell, argtype, &ap);
	va_end(ap);
	return ret;
}


CELL *libmawk_call_function(mawk_state_t *MAWK, const char *fname, const char *argtypes, ...)
{
	va_list ap;
	int numargs;
	SYMTAB *fs;
	CELL stack[EVAL_STACK_SIZE], *sp, *ret, *tmp;

	fs = mawk_find(MAWK, fname, 0);
	if ((fs == NULL) || (fs->type != ST_FUNCT)) {
		/* does not exist or not a function */
		return NULL;
	}

	sp = stack;

	va_start(ap, argtypes);
	for(numargs = 0;*argtypes != '\0';argtypes++,numargs++) {
		if (libmawk_set_cellv(MAWK, sp, *argtypes, &ap) == NULL) {
			va_end(ap);
			return NULL;
		}
		sp++;
	}
	va_end(ap);
	tmp = mawk_call(MAWK, sp-1, fs->stval.fbp, numargs);
	ret = malloc(sizeof(CELL));
	*ret = *tmp;
	return ret;
}

CELL *libmawk_call_functionp(mawk_state_t *MAWK, const char *fname, const char *argtypes, void **args)
{
	int numargs;
	SYMTAB *fs;
	CELL stack[EVAL_STACK_SIZE], *sp, *ret, *tmp;

	fs = mawk_find(MAWK, fname, 0);
	if ((fs == NULL) || (fs->type != ST_FUNCT)) {
		/* does not exist or not a function */
		return NULL;
	}

	sp = stack;

	for(numargs = 0;*argtypes != '\0';argtypes++,numargs++,args++) {
		if (libmawk_set_cellp(MAWK, sp, *argtypes, *args) == NULL)
			return NULL;
		sp++;
	}
	tmp = mawk_call(MAWK, sp-1, fs->stval.fbp, numargs);
	ret = malloc(sizeof(CELL));
	*ret = *tmp;
	return ret;
}

void libmawk_cell_destroy(mawk_state_t *MAWK, CELL *c)
{
	cell_destroy(c);
}

char *libmawk_print_cell(CELL *c, char *buff, int size)
{
	if ((buff == NULL) || (size < 1))
		return NULL;
	*buff = '\0';

	switch(c->type) {
		case C_NUM:
#warning TODO: do not use snprintf - it is not portable
#ifdef MAWK_NO_FLOAT
				snprintf(buff, size, "%d", (int)c->dval);
#else
			if (c->dval == (int)c->dval)
				snprintf(buff, size, "%d", (int)c->dval);
			else
				snprintf(buff, size, "%f", c->dval);
			break;
#endif
		case C_STRING:
			strncpy(buff, ((STRING *)(c->ptr))->str, size);
			buff[size-1] = '\0';
			break;
		
		case C_NOINIT:
		case C_STRNUM:
		case C_MBSTRN:
		case C_RE:
		case C_SPACE:
		case C_SNULL:
		case C_REPL:
		case C_REPLV:
			snprintf(buff, size, "Can't convert celltype %d\n", c->type);
			break;
		default:
			snprintf(buff, size, "Invalid celltype %d\n", c->type);
			break;
			
	}
	return buff;
}


CELL *libmawk_get_var(mawk_state_t *MAWK, const char *vname)
{
	return mawk_get_var(MAWK, vname);
}

CELL *libmawk_get_array_at(mawk_state_t *MAWK, const char *arr_name, const char *idx, int alloc)
{
	SYMTAB *fs;
	CELL idxc;

	fs = mawk_find(MAWK, arr_name, 0);

	/* does symbol exist at all? */
	if (fs == NULL)
		return NULL;

	/* exit if not an array */
	if (fs->type != ST_ARRAY)
		return NULL;

	idxc.type = C_STRING;
	idxc.ptr  = (PTR) mawk_new_STRING(MAWK, idx);

	return mawk_array_find(MAWK, (ARRAY)fs->stval.cp, &idxc, alloc);
}

int libmawk_register_function(mawk_state_t *MAWK, const char *fname, libmawk_c_function *callback)
{
	SYMTAB *sym;
	sym = mawk_find(MAWK, fname, 0);

	if (sym != NULL) {
		/* special case: we already know symbol is a function but body is empty (typical for c calls) */
		if ((sym->type == ST_FUNCT) && (sym->stval.fbp->code == NULL))
			sym->type = ST_NONE;

		/* if symbol is already defined as something else, return error */
		if (sym->type != ST_NONE)
			return 1;
	}
	else
		sym = mawk_find(MAWK, strdup(fname), 1);

	sym->type = ST_C_FUNCTION;
	sym->stval.c_function.callback = callback;
	sym->stval.c_function.userdata = MAWK->userdata;
	return 0;
}

CELL *libmawk_stackret(CELL *original_sp)
{
	return original_sp+1;
}

